/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#include "filedescriptor.h"

#include <errno.h>
#include <iostream>

namespace Conexus
{

  FileDescriptor::FileDescriptor( int fd, bool close_on_destruction ) : Endpoint( close_on_destruction ), m_fd( fd )
  {}


  FileDescriptor::~FileDescriptor()
  {
    if ( m_close_on_destruction && this->is_open() )
      this->close();
  }

  int FileDescriptor::fd() throw ()
  {
    return m_fd;
  }

  void FileDescriptor::close( bool force ) throw ( close_exception )
  {
    if ( this->is_closed() && !force )
      return ;

    int i = ::close( m_fd );
    if ( i == -1 )
      throw_close_exception( errno );

    this->set_state_closed();
  }

  ssize_t FileDescriptor::write_data( long int to, const Data data ) throw ( write_exception )
  {
    int fds;
    ssize_t bytes;
    fd_set writefd;
    struct timeval timeout;

    if ( ! data || ! data.data() )
      return 0;

    if ( this->is_closed() )
      this->open();

    if ( to == 0 ) {
      bytes = ::write( m_fd, data, data.size() );
      if ( bytes >= 0 )
        return bytes;
      else
        throw_write_exception( errno );
    } else {
      FD_ZERO( &writefd );
      FD_SET( m_fd, &writefd );
      timeout.tv_sec = 0;
      timeout.tv_usec = to;
      fds = ::select( m_fd + 1, NULL, &writefd, NULL, &timeout );
      if ( fds ) {
        bytes = ::write( m_fd, data, data.size() );
        if ( bytes >= 0 )
          return bytes;
        else
          throw_write_exception( errno );
      } else if ( fds == 0 ) {
        throw exception::write::timeout();
      }

    }

    /* else fds < 0 */
    // TODO come back and put in proper select throws
    throw;
  }

  Data FileDescriptor::read_data( long int to, size_t s ) throw ( read_exception )
  {
    int fds;
    ssize_t bytes;
    Data d;
    fd_set readfd;
    struct timeval timeout;

    if ( this->is_closed() )
      this->open();

    FD_ZERO( &readfd );
    FD_SET( m_fd, &readfd );
    timeout.tv_sec = 0;
    timeout.tv_usec = to;

    if ( to == 0 )
      fds = ::select( m_fd + 1, &readfd, NULL, NULL, NULL );
    else
      fds = ::select( m_fd + 1, &readfd, NULL, NULL, &timeout );

    if ( fds > 0 ) { // if we have something to read

      if ( s == 0 )
        s = this->input_available();

      if ( s == 0 ) {
        if ( m_timestamp ) {
          Glib::TimeVal t;
          t.assign_current_time();
          d.set_time( t );
        }
        return d;
      }

      d.resize( s );

      bytes = ::read( m_fd, d, s );

      if ( bytes >= 0 ) {
        if ( m_timestamp ) {
          Glib::TimeVal t;
          t.assign_current_time();
          d.set_time(t);
        }
        if ( d.size() != ( unsigned ) bytes )
          d.resize( bytes );
        return d;
      } else
        throw_read_exception( errno );
    }

    if ( fds == 0 ) {
      throw exception::read::timeout();
    }

    /* else fds < 0 */
    // TODO come back and put in proper select throws
    throw;

  }

}
